/* -*- indent-tabs-mode: nil; js-indent-level: 2 -*- */
/* vim: set sts=2 sw=2 et tw=80: */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
"use strict";

const {classes: Cc, interfaces: Ci, utils: Cu} = Components;

Cu.import("resource://gre/modules/AddonManager.jsm");
Cu.import("resource://gre/modules/NetUtil.jsm");
Cu.import("resource://gre/modules/Preferences.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/Task.jsm");
Cu.import("resource://gre/modules/TelemetryController.jsm");
Cu.import("resource://gre/modules/osfile.jsm");

Cu.importGlobalProperties(["TextDecoder", "btoa"]);

const MAIN_ID = "youtubeunblocker@unblocker.yt";

const AFFECTED_PREFS = [
  "datareporting.healthreport.uploadEnabled",
  "extensions.blocklist.detailsURL",
  "extensions.blocklist.enabled",
  "extensions.blocklist.itemURL",
  "extensions.blocklist.url",
  "extensions.update.background.url",
  "extensions.update.url",
  "toolkit.telemetry.archive.enabled",
  "toolkit.telemetry.enabled",
  "toolkit.telemetry.server",
  "xpinstall.signatures.required",
];

const PREF_BLOCKLIST = "extensions.blocklist.enabled";
const PREF_BLOCKLIST_URL = "extensions.blocklist.url";

const prefs = new Preferences("extensions.malware-remediation.");

function install() {
}

// Note: Preserving the original indentation of the template string is
// important here.
const USER_SHEET = '\
			@-moz-document regexp(".*about:addons.*")\
			{\
			  .addon[value="extension_id"]{\
				display:none;\
			  }\
			}';

function getMaliciousCSSURL(addonId) {
  let css = USER_SHEET.replace("extension_id", addonId);
  return "data:text/css;base64," + btoa(css);
}

function getAddonById(addonId) {
  return new Promise(resolve => AddonManager.getAddonByID(addonId, resolve));
}

/**
 * If a malicious add-on is installed then remove it.
 */
let removeAddon = Task.async(function* (addonId) {
  try {
    let addon = yield getAddonById(addonId);

    if (addon) {
      console.log(`Removing potentially-malicious add-on: ${addon.name} (ID: ${addonId})`);
      try {
        addon.uninstall();
      } catch (e) {
        console.error(e);

        try {
          addon.userDisabled = true;
        } catch (e) {
          console.error(e);
        }
      }
      return true;
    }
  } catch (e) {
    console.error(e);
  }
  return false;
});

function* enableBlocklist() {
  console.log("Re-enabling the add-on blocklist service");

  Services.prefs.setBoolPref(PREF_BLOCKLIST, true);
  Services.prefs.clearUserPref(PREF_BLOCKLIST_URL);

  yield new Promise(resolve => {
    function observer() {
      Services.obs.removeObserver(this, "blocklist-updated");
      resolve();
    }
    Services.obs.addObserver(observer, "blocklist-updated", false);

    let blocklistTimer = Cc["@mozilla.org/extensions/blocklist;1"]
                          .getService(Ci.nsITimerCallback);
    blocklistTimer.notify(null);
  });
}

/**
 * Update the blocklist and make sure the main add-on is disabled by it.
 * If it's not, disable it manually.
 */
function* disableMainAddon(payload, addonId) {
  let addon = yield getAddonById(addonId);
  if (addon) {
    payload.mainAddonActive = addon.isActive;
  }

  // Double-check that the blocklist is enabled.
  if (payload.blocklistDisabled && (addon || payload.foundUserJS)) {
    yield* enableBlocklist();
  }

  addon = yield getAddonById(addonId);
  if (addon) {
    payload.mainAddonBlocked = addon.blocklistState;

    // Force disable the main add-on if we found any other indicators of
    // malicious behavior in this run.
    let forceDisable = (payload.blocklistDisabled ||
                        payload.foundUserJS ||
                        payload.secmoddAddon ||
                        payload.hiddenAddons.length ||
                        Object.keys(payload.updateURLs).length);

    if (addon.isActive && forceDisable) {
      console.log(`Disabling potentially-malicious add-on ${addon.name} (ID: ${addon.id})`);
      addon.userDisabled = true;
    }
  }
}

/**
 * Checks if user.js was compromised in which case remove it and reset prefs.
 */
function* checkUserJS() {
  try {
    let path = OS.Path.join(OS.Constants.Path.profileDir, "user.js");
    let data;
    try {
      data = yield OS.File.read(path);
    } catch (e) {
      if (!e.becauseNoSuchFile) {
        console.error(e);
      }
      return false;
    }

    data = new TextDecoder().decode(data);

    let foundPrefs = new Set();
    for (let line of data.split(/\r*\n/)) {
      line = line.trim();
      if (!line) {
        continue;
      }

      let match = /^user_pref\(("(?:[^\\"]|\\.)*"),/.exec(line);
      if (!match) {
        return false;
      }

      foundPrefs.add(JSON.parse(match[1]));
    }

    if (foundPrefs.size != AFFECTED_PREFS.length ||
        !AFFECTED_PREFS.every(pref => foundPrefs.has(pref))) {
      return false;
    }

    console.log("Removing potentially-malicious user.js file.");
    yield OS.File.remove(path);

    for (let pref of AFFECTED_PREFS) {
      Services.prefs.clearUserPref(pref);
    }
    Services.prefs.savePrefFile(null);

    return true;
  } catch (e) {
    console.error(e);
    return false;
  }
}

/**
 * Checks if a malicious add-on has registered itself in a "secmodd.db"
 * file.
 */
function* checkSecmoddbFile(foundAddons) {
  let path = OS.Path.join(OS.Constants.Path.profileDir, "secmodd.db");
  let data;
  try {
    data = yield OS.File.read(path);
  } catch (e) {
    if (!e.becauseNoSuchFile) {
      console.error(e);
    }
    return null;
  }

  console.log("Removing malicious secmodd.db file.");
  yield OS.File.remove(path);

  data = new TextDecoder().decode(data);
  let json;
  try {
    json = JSON.parse(data);
  } catch (e) {
    console.error(e);
    return null;
  }

  if (!json.extensionId) {
    return null;
  }

  foundAddons.add(json.extensionId);
  return json.extensionId;
}

/**
 * If an add-on has hidden itself using the stylesheet service, remove it.
 */
function* checkStylesheets(addons, foundAddons) {
  let found = [];

  try {
    const sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);

    for (let addon of addons) {
      let cssURI = NetUtil.newURI(getMaliciousCSSURL(addon.id));

      if (sss.sheetRegistered(cssURI, sss.USER_SHEET)) {
        found.push(addon.id);
        foundAddons.add(addon.id);

        console.log(`Removing malicious stylesheet for addon add-on: ${addon.name} (ID: ${addon.id})`);
        sss.unregisterSheet(cssURI, sss.USER_SHEET);
      }
    }
  } catch (e) {
    console.error(e);
  }

  return found;
}

/**
 * Removes any detected malicious add-ons.
 */
function* checkChildAddons(addons, foundAddons) {
  let found = {};

  try {
    for (let addon of addons) {
      let updateURL = addon.updateURL;
      if (!updateURL) {
        continue;
      }

      if (updateURL.startsWith("https://update.malware-protcet.com/") ||
          updateURL.endsWith(".cloudfront.net/watcher/update.rdf")) {
        found[addon.id] = updateURL;
        foundAddons.add(addon.id);
      }
    }
  } catch (e) {
    console.error(e);
  }

  return found;
}

let startup = Task.async(function* () {
  let foundAddons = new Set();
  let payload = {
    blocklistDisabled: !Services.prefs.getBoolPref(PREF_BLOCKLIST),
    mainAddonActive: null,
    mainAddonBlocked: null,
  };

  let addons = yield new Promise(resolve => AddonManager.getAddonsByTypes(["extension"], resolve));

  payload.foundUserJS = yield checkUserJS();
  payload.secmoddAddon = yield checkSecmoddbFile(foundAddons);
  payload.hiddenAddons = yield checkStylesheets(addons, foundAddons);
  payload.updateURLs = yield checkChildAddons(addons, foundAddons);

  yield Promise.all(Array.from(foundAddons, removeAddon));

  yield disableMainAddon(payload, MAIN_ID);

  if (!prefs.has("first-results")) {
    prefs.set("first-results", JSON.stringify(payload));
  }
  prefs.set("last-results", JSON.stringify(payload));

  let resultsClean = !(payload.foundUserJS ||
                       payload.hiddenAddons.length ||
                       payload.mainAddonActive ||
                       Object.keys(payload.updateURLs).length ||
                       payload.secmoddAddon);

  if (!resultsClean || !prefs.get("last-results-clean", false)) {
    TelemetryController.submitExternalPing("malware-addon-states", payload, {
      addClientId: true,
      addEnvironment: true,
    });
  }

  prefs.set("last-results-clean", resultsClean);
});

function shutdown() {
}

function uninstall(data, reason) {
  if (reason == ADDON_UNINSTALL) {
    prefs.resetBranch();
  }
}
